/*
 *
 * (C) 2004 - Luca Deri <deri@ntop.org>
 *
 */

#include <linux/version.h>
#include <linux/config.h>
#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/socket.h>
#include <linux/skbuff.h>
#include <linux/rtnetlink.h>
#include <linux/in.h>
#include <linux/in6.h>
#include <linux/init.h>
#include <linux/filter.h>
#include <linux/ring.h>
#if (LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0))
#include <net/xfrm.h>
#else
#include <linux/poll.h>
#endif
#include <net/sock.h>
#include <asm/io.h>        /* needed for virt_to_phys()    */

/* #define RING_DEBUG */

/* ************************************************* */

struct ring_list {
  struct sock      *sk;
  struct ring_list *next;
};

struct ring_opt {
  struct net_device *ring_netdev;

  /* Packet buffers */
  FlowSlotInfo *slots_info;
  char **pageBuffers; /* pointer to slots_info (recasted to char*) */
  u_int tot_num_pages, pkts_per_page;

  /* Packet Sampling */
  u_int pktToSample, sample_rate;

  /* BPF Filter */
  struct sk_filter *bpfFilter;
  /* struct sock_fprog fprog; */

  /* Locks */
  atomic_t num_ring_slots_waiters;
  wait_queue_head_t ring_slots_waitqueue;
  rwlock_t ring_index_lock;

  /* Indexes (Internal) */
  u_int insert_page_id, insert_slot_id, insert_pkts_per_page;
};

/* ************************************************* */

/* List of all ring sockets. */
static struct ring_list *ring_table;
static rwlock_t ring_mgmt_lock = RW_LOCK_UNLOCKED;

/* ********************************** */

/* Forward */
static struct proto_ops ring_ops;
static int my_ring_handler(struct sk_buff *skb);

/* Extern */

/* ********************************** */

/* Defaults */
static u_int bucket_len = 128, num_slots = 4096, sample_rate = 1, transparent_mode = 0;

MODULE_PARM(bucket_len, "i");
MODULE_PARM_DESC(bucket_len, "Number of ring buckets");
MODULE_PARM(num_slots,  "i");
MODULE_PARM_DESC(num_slots,  "Number of ring slots");
MODULE_PARM(sample_rate, "i");
MODULE_PARM_DESC(sample_rate, "Ring packet sample rate");
MODULE_PARM(transparent_mode, "i");
MODULE_PARM_DESC(transparent_mode, "Set to 1 to set transparent mode (slower but backwards compatible)");

/* ********************************** */

#if (LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0))
#define ring_sk(__sk) ((struct ring_opt *)(__sk)->sk_protinfo)
#else
#define ring_sk(__sk) ((__sk)->protinfo.pf_ring)
#endif
#define _rdtsc()      ({ uint64_t x; asm volatile("rdtsc" : "=A" (x)); x; })

/* ********************************** */

static void* allocate_page(void) {
  struct page *page;
  void *ptr = kmalloc(PAGE_SIZE, GFP_ATOMIC);

  if(ptr == NULL) return(NULL);
  memset(ptr, 0, PAGE_SIZE);

  for(page = virt_to_page(ptr); page < virt_to_page(ptr+PAGE_SIZE); page++)
    SetPageReserved(page);

  return(ptr);
}

/* ********************************** */

static void free_allocated_page(void *ptr) {
  struct page *page;

  for(page = virt_to_page(ptr); page < virt_to_page(ptr+PAGE_SIZE); page++)
    ClearPageReserved(page);

  kfree(ptr);
}

/* ********************************** */

static void ring_sock_destruct(struct sock *sk) {

#if (LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0))
  skb_queue_purge(&sk->sk_receive_queue);

  if (!sock_flag(sk, SOCK_DEAD)) {
#ifdef RING_DEBUG
    printk("Attempt to release alive ring socket: %p\n", sk);
#endif
    return;
  }

  BUG_TRAP(!atomic_read(&sk->sk_rmem_alloc));
  BUG_TRAP(!atomic_read(&sk->sk_wmem_alloc));
#else

  BUG_TRAP(atomic_read(&sk->rmem_alloc)==0);
  BUG_TRAP(atomic_read(&sk->wmem_alloc)==0);

  if (!sk->dead) {
#ifdef RING_DEBUG
    printk("Attempt to release alive ring socket: %p\n", sk);
#endif
    return;
  }
#endif

  kfree(ring_sk(sk));

#if (LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0))
  MOD_DEC_USE_COUNT;
#endif
}

/* ********************************** */

static void ring_insert(struct sock *sk) {
  struct ring_list *next;

#ifdef RING_DEBUG
  printk("RING: ring_insert()\n");
#endif

  write_lock(&ring_mgmt_lock);
  next = kmalloc(sizeof(struct ring_list), GFP_ATOMIC);
  if(next != NULL) {
    next->sk = sk;
    next->next = ring_table;
    ring_table = next;
  }

  write_unlock(&ring_mgmt_lock);
}

/* ********************************** */

static void ring_remove(struct sock *sk) {
  struct ring_list *ptr, *prev = NULL;

  ptr = ring_table;

  if(ptr != NULL) {
    if(ptr->sk == sk) {
      struct ring_list *next = ptr->next;
      kfree(ptr);

      if(prev != NULL)
	prev->next = next;
      else
	ring_table = next;
    } else {
      prev = ptr;
      ptr = ptr->next;
    }
  }
}

/* ********************************** */

static FlowSlot* get_insert_slot(struct ring_opt *pfr) {
  char *page = pfr->pageBuffers[pfr->insert_page_id];

#ifdef RING_DEBUG
  printk("get_insert_slot(%d): page=%d slot=%d\n",
	 pfr->slots_info->insert_idx,
	 pfr->insert_page_id, pfr->insert_slot_id);
#endif

  return((FlowSlot*)&page[pfr->insert_slot_id]);
}

/* ********************************** */

static FlowSlot* get_remove_slot(struct ring_opt *pfr) {
  u_short pageId = (pfr->slots_info->remove_idx/pfr->slots_info->pkts_per_page)+1;
  u_short slotId = (pfr->slots_info->remove_idx % pfr->slots_info->pkts_per_page)*pfr->slots_info->slot_len;
  char *page = pfr->pageBuffers[pageId];

#ifdef RING_DEBUG
  printk("get_remove_slot(%d): page=%d slot=%d\n",
	 pfr->slots_info->remove_idx, pageId, slotId);
#endif

  return((FlowSlot*)&page[slotId]);
}

/* ********************************** */

static void add_skb_to_ring(struct sk_buff *skb, struct ring_opt *pfr) {
  FlowSlot *theSlot;
  int idx;
  
  write_lock(&pfr->ring_index_lock);

  /* BPF Filtering (from af_packet.c) */
  if(pfr->bpfFilter != NULL) {
    unsigned res = 1, len;

    len = skb->len-skb->data_len;
    
    skb->data -= SKB_DISPLACEMENT;
    res = sk_run_filter(skb, pfr->bpfFilter->insns, pfr->bpfFilter->len);
    skb->data += SKB_DISPLACEMENT;

    if(res == 0) {
      /* Filter failed */
      write_unlock(&pfr->ring_index_lock);
      return;
    }
  }

  /* ************************** */

  pfr->slots_info->tot_pkts++;

  if(pfr->sample_rate > 1) {
    if(pfr->pktToSample == 0)
      pfr->pktToSample = pfr->sample_rate;
    else {
      pfr->pktToSample--;
      write_unlock(&pfr->ring_index_lock);
      return;
    }
  }

#ifdef RING_DEBUG
  printk("add_skb_to_ring(skb) [len=%d][tot=%d][insertIdx=%d]\n",
	 (int)skb->len, pfr->slots_info->tot_pkts,
	 pfr->slots_info->insert_idx);
#endif

  idx = pfr->slots_info->insert_idx;
  theSlot = get_insert_slot(pfr);

  if(theSlot->slot_state == 0) {
    struct pcap_pkthdr *hdr;
    unsigned int bucketSpace;
    char *bucket;

    idx++;
    if(idx == pfr->slots_info->tot_slots) {
      pfr->slots_info->insert_idx = 0;
      pfr->insert_page_id = 1, pfr->insert_slot_id = 0, pfr->insert_pkts_per_page = 0;
    } else {
      pfr->slots_info->insert_idx = idx;

      pfr->insert_pkts_per_page++, pfr->insert_slot_id += pfr->slots_info->slot_len;

      if(pfr->insert_pkts_per_page >= pfr->slots_info->pkts_per_page) {
	pfr->insert_page_id++, pfr->insert_slot_id = 0, pfr->insert_pkts_per_page = 0;
      }
    }

    write_unlock(&pfr->ring_index_lock);

    bucketSpace = pfr->slots_info->slot_len 
      - sizeof(u_char)  /* flowSlot.slot_state */
      - sizeof(struct pcap_pkthdr)
      - 10 /* safe boundary */;

    bucket = &theSlot->bucket;
    hdr = (struct pcap_pkthdr*)bucket;

    if(skb->stamp.tv_sec == 0) do_gettimeofday(&skb->stamp);

    hdr->ts.tv_sec = skb->stamp.tv_sec, hdr->ts.tv_usec = skb->stamp.tv_usec;
    hdr->caplen    = skb->len+SKB_DISPLACEMENT;

    if(hdr->caplen > bucketSpace)
      hdr->caplen = bucketSpace;

    hdr->len = skb->len+SKB_DISPLACEMENT;
    memcpy(&bucket[sizeof(struct pcap_pkthdr)], skb->data-SKB_DISPLACEMENT, hdr->caplen);

#ifdef RING_DEBUG
    printk("add_skb_to_ring(%d): [bucketSpace=%d]"
	   "[hdr.caplen=%d][skb->len=%d]"
	   "[pcap_pkthdr=%d][slotState@removeIdx(%d)=%d]\n",
	   idx-1, bucketSpace, hdr->caplen, skb->len,
	   sizeof(struct pcap_pkthdr),
	   pfr->slots_info->remove_idx,
	   get_remove_slot(pfr)->slot_state);
#endif

    theSlot->slot_state = 1;
  } else {
    pfr->slots_info->tot_lost++;
    write_unlock(&pfr->ring_index_lock);

#ifdef RING_DEBUG
    printk("add_skb_to_ring(skb): packet lost [loss=%d]\n",
	   pfr->slots_info->tot_lost);
#endif
  }

  /* wakeup in case of poll() */
  if(waitqueue_active(&pfr->ring_slots_waitqueue))
    wake_up_interruptible(&pfr->ring_slots_waitqueue);
}

/* ********************************** */

static int my_ring_handler(struct sk_buff *skb) {
  struct sock *skElement;
  int rc = 0;
#ifdef RING_DEBUG
  int num=0;
#endif
  struct ring_list *ptr;
#ifdef PROFILING
  uint64_t rdt = _rdtsc(), rdt1, rdt2;
#endif
  if(!skb) return(0);

#ifdef RING_DEBUG
    printk("my_ring_handler() [len=%d][dev=%s]\n", skb->len,
	   skb->dev->name == NULL ? "<NULL>" : skb->dev->name);
#endif

  read_lock(&ring_mgmt_lock);

#ifdef PROFILING
  rdt1 = _rdtsc();
#endif

  ptr = ring_table;
  while(ptr != NULL) {
    struct ring_opt *pfr;

    skElement = ptr->sk;

    pfr = ring_sk(skElement);
    if(pfr->ring_netdev == skb->dev) {
      /* We've found the ring where the packet can be stored */
      add_skb_to_ring(skb, pfr);

      /* DO NOT DISABLE THE MAIN NETWORK INTERFACE !!!! */
      rc = 1; /* Ring found: we've done our job */
    }

    ptr = ptr->next;
#ifdef RING_DEBUG
    num++;
#endif
  }

#ifdef RING_DEBUG
  /* printk("# loops: %d\n", num); */
#endif

#ifdef PROFILING
  rdt1 = _rdtsc()-rdt1;
#endif

  read_unlock(&ring_mgmt_lock);

#ifdef PROFILING
  rdt2 = _rdtsc();
#endif

  if(transparent_mode) rc = 0;

  if(rc != 0)
    dev_kfree_skb(skb); /* Free the skb */

#ifdef PROFILING
  rdt2 = _rdtsc()-rdt2;
  rdt = _rdtsc()-rdt;

#ifdef RING_DEBUG
  printk("# cycles: %d [lock costed %d %d%%][free costed %d %d%%]\n",
	 (int)rdt, rdt-rdt1,
	 (int)((float)((rdt-rdt1)*100)/(float)rdt),
	 rdt2,
	 (int)((float)(rdt2*100)/(float)rdt));
#endif
#endif

  return(rc); /*  0 = packet not handled */
}

/* ********************************** */

static int ring_create(struct socket *sock, int protocol) {
  struct sock *sk;
  struct ring_opt *pfr;
  int err;

#ifdef RING_DEBUG
  printk("RING: ring_create()\n");
#endif

  /* Are you root, superuser or so ? */
  if(!capable(CAP_NET_ADMIN))
    return -EPERM;

  if(sock->type != SOCK_RAW)
    return -ESOCKTNOSUPPORT;

  if(protocol != htons(ETH_P_ALL))
    return -EPROTONOSUPPORT;

#if (LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0))
  MOD_INC_USE_COUNT;
#endif

  err = -ENOMEM;
  sk = sk_alloc(PF_RING, GFP_KERNEL, 1
#if (LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0))
		, NULL
#endif
		);
  if (sk == NULL)
    goto out;

  sock->ops = &ring_ops;
  sock_init_data(sock, sk);
#if (LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0))
  sk_set_owner(sk, THIS_MODULE);
#endif

  err = -ENOMEM;
  pfr = ring_sk(sk) = kmalloc(sizeof(*pfr), GFP_KERNEL);
  if (!pfr) {
    sk_free(sk);
    goto out;
  }
  memset(pfr, 0, sizeof(*pfr));
  init_waitqueue_head(&pfr->ring_slots_waitqueue);
  pfr->ring_index_lock = RW_LOCK_UNLOCKED;
  atomic_set(&pfr->num_ring_slots_waiters, 0);
#if (LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0))
  sk->sk_family       = PF_RING;
  sk->sk_destruct     = ring_sock_destruct;
#else
  sk->family          = PF_RING;
  sk->destruct        = ring_sock_destruct;
  sk->num             = protocol;
#endif

  ring_insert(sk);
  return(0);
 out:
#if (LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0))
  MOD_DEC_USE_COUNT;
#endif
  return err;
}

static int ring_release(struct socket *sock)
{
  struct sock *sk = sock->sk;
  struct ring_opt *pfr = ring_sk(sk);

  if(!sk)
    return 0;

#ifdef RING_DEBUG
  printk("RING: ring_release()\n");
#endif

  write_lock_bh(&ring_mgmt_lock);
  ring_remove(sk);

  sock_orphan(sk);
  sock->sk = NULL;

  /* Free the ring buffer */
  if(pfr->pageBuffers) {
    int i;

    for(i=0; i<pfr->tot_num_pages; i++)
      free_allocated_page(pfr->pageBuffers[i]);

    kfree(pfr->pageBuffers);
  }

  kfree(pfr);
  ring_sk(sk) = NULL;

#if (LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0))
  skb_queue_purge(&sk->sk_write_queue);
#endif
  sock_put(sk);

  write_unlock_bh(&ring_mgmt_lock);

  return 0;
}
/* ********************************** */
/*
 * We create a ring for this socket and bind it to the specified device
 */
static int packet_ring_bind(struct sock *sk, struct net_device *dev)
{
  u_int the_slot_len, i;
  struct ring_opt *pfr = ring_sk(sk);

  if(!dev) return(-1);

#ifdef RING_DEBUG
  printk("RING: packet_ring_bind(%s) called\n", dev->name);
#endif

  /* **********************************************

  *************************************
  *                                   *
  *        FlowSlotInfo               *
  *                                   *
  ************************************* <-+
  *        FlowSlot                   *   |
  *************************************   |
  *        FlowSlot                   *   |
  *************************************   +- num_slots
  *        FlowSlot                   *   |
  *************************************   |
  *        FlowSlot                   *   |
  ************************************* <-+

  pages
  [0] Reserved for FlowSlotInfo
  [1]
  ..
  [tot_num_pages-1] are reserved for packets

  ********************************************** */

  the_slot_len = sizeof(u_char)  /* flowSlot.slot_state */
               + sizeof(u_short) /* flowSlot.slot_len   */
               + bucket_len      /* flowSlot.bucket     */; 
  pfr->pkts_per_page = PAGE_SIZE/the_slot_len;

  /* 1 page only */
  if(pfr->pkts_per_page > num_slots) {
    pfr->pkts_per_page = num_slots;
    pfr->tot_num_pages = 2; /* 1 page + 1 page header */
  } else {
    /* 1 is for FlowSlotInfo header */;
    pfr->tot_num_pages = 1+((num_slots+pfr->pkts_per_page-1)
			    /pfr->pkts_per_page);
  }

#ifdef RING_DEBUG
  printk("RING: %d pages to allocate\n", pfr->tot_num_pages);
#endif
  pfr->pageBuffers = kmalloc(sizeof(char*)*pfr->tot_num_pages, GFP_KERNEL);

  if(pfr->pageBuffers == NULL) {
#ifdef RING_DEBUG
    printk("ERROR: not enough memory while allocating pageBuffers\n");
#endif
    return(-1);
  }

  for(i=0; i<pfr->tot_num_pages; i++) {
    pfr->pageBuffers[i] = allocate_page();
    if(pfr->pageBuffers[i] == NULL) {
      int j;

      for(j=0; j<i; j++) free_allocated_page(pfr->pageBuffers[j]);
      kfree(pfr->pageBuffers);
#ifdef RING_DEBUG
      printk("ERROR: not enough memory [page=%d]\n", i);
#endif
      return(-1);
    }
  }

  pfr->slots_info = (FlowSlotInfo*)pfr->pageBuffers[0];

  pfr->slots_info->slot_len      = the_slot_len;
  pfr->slots_info->tot_slots     = num_slots;
  pfr->slots_info->sample_rate   = sample_rate;
  pfr->slots_info->tot_num_pages = pfr->tot_num_pages;
  pfr->slots_info->pkts_per_page = pfr->pkts_per_page;

#ifdef RING_DEBUG
  printk("RING: allocated %d slots [slot_len=%d]"
	 "[tot_slots=%d][pages=%d][pkt x page=%d]\n",
	 num_slots,
	 pfr->slots_info->slot_len, pfr->slots_info->tot_slots,
	 pfr->slots_info->tot_num_pages,
	 pfr->slots_info->pkts_per_page);
#endif

  pfr->insert_page_id = 1, pfr->insert_slot_id = 0,
    pfr->insert_pkts_per_page = 0;

  /*
    IMPORTANT
    Leave this statement here as last one. In fact when
    the ring_netdev != NULL the socket is ready to be used.
  */
  pfr->ring_netdev = dev;

  return(0);
}

/* ************************************* */

/* Bind to a device */
static int ring_bind(struct socket *sock, struct sockaddr *sa, int addr_len)
{
  struct sock *sk=sock->sk;
  struct net_device *dev = NULL;

#ifdef RING_DEBUG
  printk("RING: ring_bind() called\n");
#endif

  /*
   *	Check legality
   */
  if (addr_len != sizeof(struct sockaddr))
    return -EINVAL;
  if (sa->sa_family != PF_RING)
    return -EINVAL;

  /* Safety check: add trailing zero if missing */
  sa->sa_data[sizeof(sa->sa_data)-1] = '\0';

#ifdef RING_DEBUG
  printk("RING: searching device %s\n", sa->sa_data);
#endif

  if((dev = __dev_get_by_name(sa->sa_data)) == NULL) {
#ifdef RING_DEBUG
    printk("RING: search failed\n");
#endif
    return(-EINVAL);
  } else
    return(packet_ring_bind(sk, dev));
}

/* ************************************* */

#if (LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0))
volatile void* virt_to_kseg(volatile void* address) {
  pte_t *pte;

  pte = pte_offset_map(pmd_offset(pgd_offset_k((unsigned long) address),
				  (unsigned long) address),
		       (unsigned long) address);
  return((volatile void*)pte_page(*pte));
}
#else /* 2.4 */

/* http://www.scs.ch/~frey/linux/memorymap.html */
volatile void *virt_to_kseg(volatile void *address)
{
  pgd_t *pgd; pmd_t *pmd; pte_t *ptep, pte;
  unsigned long va, ret = 0UL;

  va=VMALLOC_VMADDR((unsigned long)address);

  /* get the page directory. Use the kernel memory map. */
  pgd = pgd_offset_k(va);

  /* check whether we found an entry */
  if (!pgd_none(*pgd))
    {
      /* get the page middle directory */
      pmd = pmd_offset(pgd, va);
      /* check whether we found an entry */
      if (!pmd_none(*pmd))
	{
	  /* get a pointer to the page table entry */
	  ptep = pte_offset(pmd, va);
	  pte = *ptep;
	  /* check for a valid page */
	  if (pte_present(pte))
	    {
	      /* get the address the page is refering to */
	      ret = (unsigned long)page_address(pte_page(pte));
	      /* add the offset within the page to the page address */
	      ret |= (va & (PAGE_SIZE -1));
	    }
	}
    }
  return((volatile void *)ret);
}
#endif

/* ************************************* */

static int ring_mmap(struct file *file,
		     struct socket *sock,
		     struct vm_area_struct *vma)
{
  struct sock *sk = sock->sk;
  struct ring_opt *po = ring_sk(sk);
  unsigned long size, start;
  u_int pagesToMap, i;

  if(!po->pageBuffers) {
#ifdef RING_DEBUG
    printk("RING: ring_mmap() failed: mapping area to an unbound socket\n");
#endif
    return -EINVAL;
  }

  size = (unsigned long)(vma->vm_end-vma->vm_start);

  if(size % PAGE_SIZE) {
#ifdef RING_DEBUG
    printk("RING: ring_mmap() failed: len is not multiple of PAGE_SIZE\n");
#endif
    return -EINVAL;
  }

  pagesToMap = size/PAGE_SIZE;

  /* if userspace tries to mmap beyond end of our buffer, fail */
  if(pagesToMap > po->tot_num_pages) {
#ifdef RING_DEBUG
    printk("proc_mmap() failed: area too large [%d > %d]\n",
	   pagesToMap, po->tot_num_pages);
#endif
    return(-EINVAL);
  }

#ifdef RING_DEBUG
  printk("RING: mmap [slot_len=%d][tot_slots=%d] for ring on device %s\n",
	 po->slots_info->slot_len, po->slots_info->tot_slots,
	 po->ring_netdev->name);
#endif

  /* we do not want to have this area swapped out, lock it */
  vma->vm_flags |= VM_LOCKED;
  start = vma->vm_start;

  for(i=0; i<pagesToMap; i++) {
    if(remap_page_range(
#if (LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0))
			vma,
#endif
			start,
			virt_to_phys(po->pageBuffers[i]),
			PAGE_SIZE, vma->vm_page_prot)) {
#ifdef RING_DEBUG
      printk("remap_page_range() failed\n");
#endif
      return -EAGAIN;
    }

    start += PAGE_SIZE;
  }

#ifdef RING_DEBUG
  printk("proc_mmap(pagesToMap=%d): success.\n", pagesToMap);
#endif

  return 0;
}

/* ************************************* */

unsigned int ring_poll(struct file * file,
		       struct socket *sock, poll_table *wait)
{
  FlowSlot* slot;
  struct ring_opt *pfr = ring_sk(sock->sk);

#ifdef RING_DEBUG
  printk("poll called\n");
#endif

  slot = get_remove_slot(pfr);

  if(slot->slot_state == 0)
    poll_wait(file, &pfr->ring_slots_waitqueue, wait);

#ifdef RING_DEBUG
  printk("poll returning %d\n", slot->slot_state);
#endif

  if(slot->slot_state == 1)
    return(POLLIN | POLLRDNORM);
  else
    return 0;
}

/* ************************************* */

/* Code taken/inspired from core/sock.c */
static int ring_setsockopt(struct socket *sock, int level, int optname,
			   char *optval, int optlen)
{
  struct ring_opt *pfr = ring_sk(sock->sk);
  int val, found = 0, ret = 0;

  if(optlen<sizeof(int))
    return(-EINVAL);

  if (get_user(val, (int *)optval))
    return -EFAULT;

  write_lock(&ring_mgmt_lock);

  switch(optname)
    {
    case SO_ATTACH_FILTER:
      found = 1;
      ret = -EINVAL;
      if (optlen == sizeof(struct sock_fprog)) {
	unsigned int fsize;
	struct sock_fprog fprog;

	ret = -EFAULT;
	if(copy_from_user(&fprog, optval, sizeof(fprog)))
	  break;

	fsize = sizeof(struct sock_filter) * fprog.len;
	pfr->bpfFilter = kmalloc(fsize, GFP_KERNEL);

	if(pfr->bpfFilter == NULL) {
	  ret = -ENOMEM;
	  break;
	}

	if(copy_from_user(pfr->bpfFilter->insns, fprog.filter, fsize))
	  break;
	
	pfr->bpfFilter->len = fprog.len;

	if(sk_chk_filter(pfr->bpfFilter->insns, pfr->bpfFilter->len) != 0) {
	  /* Bad filter specified */
	  kfree(pfr->bpfFilter);
	  pfr->bpfFilter = NULL;
	  break;
	}
      }
      ret = 0;
      break;

    case SO_DETACH_FILTER:
      found = 1;
      if(pfr->bpfFilter != NULL) {
	kfree(pfr->bpfFilter);
	pfr->bpfFilter = NULL;
	break;
      }
      ret = -ENONET;
      break;
    }

  write_unlock(&ring_mgmt_lock);

  if(found)
    return ret;
  else
    return(sock_setsockopt(sock, level, optname, optval, optlen));
}

/* ************************************* */

static int ring_ioctl(struct socket *sock, unsigned int cmd, unsigned long arg)
{

  switch(cmd) 
    {
    case SIOCGIFFLAGS:
    case SIOCSIFFLAGS:
    case SIOCGIFCONF:
    case SIOCGIFMETRIC:
    case SIOCSIFMETRIC:
    case SIOCGIFMEM:
    case SIOCSIFMEM:
    case SIOCGIFMTU:
    case SIOCSIFMTU:
    case SIOCSIFLINK:
    case SIOCGIFHWADDR:
    case SIOCSIFHWADDR:
    case SIOCSIFMAP:
    case SIOCGIFMAP:
    case SIOCSIFSLAVE:
    case SIOCGIFSLAVE:
    case SIOCGIFINDEX:
    case SIOCGIFNAME:
    case SIOCGIFCOUNT:
    case SIOCSIFHWBROADCAST:
      return(dev_ioctl(cmd,(void *) arg));

    default:
      return -EOPNOTSUPP;
    }

  return 0;
}

/* ************************************* */

static struct proto_ops ring_ops = {
  .family	=	PF_RING,
#if (LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0))
  .owner	=	THIS_MODULE,
#endif

  /* Operations that make no sense on ring sockets. */
  .connect	=	sock_no_connect,
  .socketpair	=	sock_no_socketpair,
  .accept	=	sock_no_accept,
  .getname	=	sock_no_getname,
  .listen	=	sock_no_listen,
  .shutdown	=	sock_no_shutdown,
  .sendpage	=	sock_no_sendpage,
  .sendmsg	=	sock_no_sendmsg,
  .recvmsg	=	sock_no_recvmsg,
  .getsockopt	=	sock_no_getsockopt,

  /* Now the operations that really occur. */
  .release	=	ring_release,
  .bind		=	ring_bind,
  .mmap		=	ring_mmap,
  .poll		=	ring_poll,
  .setsockopt	=	ring_setsockopt,
  .ioctl	=	ring_ioctl,
};

static struct net_proto_family ring_family_ops = {
  .family	=	PF_RING,
  .create	=	ring_create,
#if (LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0))
  .owner	=	THIS_MODULE,
#endif
};

static void __exit ring_exit(void)
{
  while(ring_table != NULL) {
    struct ring_list *next = ring_table->next;
    kfree(ring_table);
    ring_table = next;
  }

  set_ring_handler(NULL);
  sock_unregister(PF_RING);
}

static int __init ring_init(void)
{
  ring_table = NULL;
  sock_register(&ring_family_ops);
  set_ring_handler(my_ring_handler);
  return 0;
}

module_init(ring_init);
module_exit(ring_exit);
MODULE_LICENSE("GPL");

#if (LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0))
MODULE_ALIAS_NETPROTO(PF_RING);
#endif
